==> TundereBird-0.24.tar.bz2: TundereBird-0.24/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Protocol.Client/ImapResponseReceiver.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

using Octets = Smdn.Formats.Octets;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Protocol.Client {
  public class ImapResponseReceiver : ImapReceiver {
    public bool ResponseContinuing {
      get { return dataResponseContinuing; }
    }

    public bool ReceiveResponseAsUTF8 {
      get; set;
    }

    public ImapResponseReceiver(LineOrientedBufferedStream stream)
      : base(stream)
    {
      this.fragmentedResponseBuffer = new ByteStringBuilder(stream.BufferSize);

      ReceiveResponseAsUTF8 = false;
    }

    private ByteStringBuilder fragmentedResponseBuffer;
    private int requiredDataLengthAtLeast = 0;
    private bool dataResponseContinuing = false;

    public ImapResponse ReceiveResponse()
    {
      var line = Receive(requiredDataLengthAtLeast);

      fragmentedResponseBuffer.Append(line);

      if (fragmentedResponseBuffer.Length < requiredDataLengthAtLeast)
        return null; // need to read more

      try {
        ByteString notParsedLines = null;

        var response = (dataResponseContinuing)
          ? ParseDataResponse(fragmentedResponseBuffer.ToByteString(), out notParsedLines, out requiredDataLengthAtLeast)
          : ParseResponce(fragmentedResponseBuffer.ToByteString(), out notParsedLines, out requiredDataLengthAtLeast);

        fragmentedResponseBuffer.Length = 0; // discard buffered

        if (notParsedLines != null)
          fragmentedResponseBuffer.Append(notParsedLines);

        return response;
      }
      catch (Exception ex) {
        fragmentedResponseBuffer.Length = 0; // discard buffered
        if (ex is ImapMalformedDataException)
          throw new ImapMalformedResponseException(ex.Message);
        else
          throw ex;
      }
    }

    private ImapResponse ParseResponce(ByteString line, out ByteString notParsedLines, out int requiredBytes)
    {
      notParsedLines = null;
      requiredBytes = 0;

      // response        = *(continue-req / response-data) response-done

      // continue-req    = "+" SP (resp-text / base64) CRLF
      if (line.StartsWith("+ "))
        return new ImapCommandContinuationRequest(line.Substring(2, line.Length - 4).ToString()); // remove leading "+" SP and trailing CRLF

      // greeting        = "*" SP (resp-cond-auth / resp-cond-bye) CRLF
      // response-done   = response-tagged / response-fatal
      // response-data   = "*" SP (resp-cond-state / resp-cond-bye /
      //                   mailbox-data / message-data / capability-data) CRLF
      // response-fatal  = "*" SP resp-cond-bye CRLF
      //                     ; Server closes connection immediately
      // response-tagged = tag SP resp-cond-state CRLF

      // ("*" / tag) SP
      var tagSep = line.IndexOf(Octets.SP);

      if (tagSep == -1)
        // response-done and response-data must contain SP
        throw new ImapMalformedResponseException("malformed response-done/response-data", line.ToString());

      var tag = line.Substring(0, tagSep);
      var untagged = tag.Equals("*");

      // ("OK" / "BAD" / "NO" / "BYE" / "PREAUTH" / text) SP
      var respCondSep = line.IndexOf(Octets.SP, tagSep + 1);
      ImapResponseCondition? cond = null;

      if (respCondSep == -1) {
        if (untagged)
          // '* SEARCH\r\n' (mailbox-data which contains no SP)
          cond = null;
        else
          throw new ImapMalformedResponseException("malformed response-data", line.ToString());
      }
      else {
        cond = ParseCondition(line.Substring(tagSep + 1, respCondSep - tagSep - 1));
      }

      if (cond != null || line.StartsWith("* [")) {
        // resp-cond-auth / resp-cond-state / resp-cond-bye
        ByteString respText;

        if (cond == null)
          respText = line.Substring(tagSep + 1);
        else
          respText = line.Substring(respCondSep + 1);

        var responseText = ParseRespText(respText);

        if (untagged)
          return new ImapUntaggedStatusResponse((ImapResponseCondition)cond,
                                            responseText);
        else
          return new ImapTaggedStatusResponse(tag.ToString(),
                                          (ImapResponseCondition)cond,
                                          responseText);
      }

      // mailbox-data / message-data / capability-data etc.
      return ParseDataResponse(line, out notParsedLines, out requiredBytes);
    }

    private static readonly Dictionary<ByteString, ImapResponseCondition> conditions = new Dictionary<ByteString, ImapResponseCondition> {
      {new ByteString("OK"),      ImapResponseCondition.Ok},
      {new ByteString("NO"),      ImapResponseCondition.No},
      {new ByteString("BAD"),     ImapResponseCondition.Bad},
      {new ByteString("BYE"),     ImapResponseCondition.Bye},
      {new ByteString("PREAUTH"), ImapResponseCondition.PreAuth},
    };

    private ImapResponseCondition? ParseCondition(ByteString cond)
    {
      ImapResponseCondition c;

      if (conditions.TryGetValue(cond.ToUpper(), out c))
        return c;
      else
        return null;
    }

    private ImapResponseText ParseRespText(ByteString respText)
    {
      // resp-text       = ["[" resp-text-code "]" SP] text
      if (respText[0] != '[')
        // no resp-text-code
        return new ImapResponseText(DecodeRespTextString(respText.Substring(0, respText.Length - 2))); // remove trailing CRLF

      // exists resp-text-code
      var respTextCodeEnd = respText.IndexOf(']');
      var respTextCode    = respText.Substring(1, respTextCodeEnd - 1);
      var codeSep         = respTextCode.IndexOf(Octets.SP);
      var text            = ByteString.CreateEmpty();

      if (respTextCodeEnd < respText.Length) {
        text = respText.Substring(respTextCodeEnd + 1);
        text = text.Trim(); // remove SP, CRLF
      }

      if (codeSep == -1)
        // no arguments; "READ-ONLY" / "READ-WRITE" / etc...
        return new ImapResponseText(ImapResponseCode.GetKnownOrCreate(respTextCode.ToString()),
                                    new ImapData[] {},
                                    DecodeRespTextString(text));

      // exist arguments; "UIDNEXT" SP nz-number / "UIDVALIDITY" SP nz-number / etc...
      var codeString  = respTextCode.Substring(0, codeSep);
      var reader = new DataReader(respTextCode, codeSep + 1);

      ImapData[] arguments;
      int discard = 0;

      if (!ParseData(reader, out arguments, ref discard))
        throw new ImapMalformedResponseException("malformed resp-text-code", respTextCode.Substring(codeSep + 1).ToString());

      return new ImapResponseText(ImapResponseCode.GetKnownOrCreate(codeString.ToString()),
                                  arguments,
                                  DecodeRespTextString(text));
    }

    private string DecodeRespTextString(ByteString text)
    {
      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 3.5. Formal Syntax
       *     resp-text         = ["[" resp-text-code "]" SP ] UTF8-TEXT-CHAR
       *                         *(UTF8-TEXT-CHAR / "[")
       *         ; After the server is changed to a language other than
       *         ; i-default, this resp-text rule replaces the resp-text
       *         ; rule from [RFC3501].
       */
      if (ReceiveResponseAsUTF8)
        return Encoding.UTF8.GetString(text.ByteArray);
      else
        return text.ToString();
    }

    private ImapDataResponse ParseDataResponse(ByteString lines, out ByteString notParsedLines, out int requiredBytes)
    {
      notParsedLines = null;
      requiredBytes = 0;

      // mailbox-data / message-data / capability-data etc.

      // ("*" / tag) SP
      var tagSep = lines.IndexOf(Octets.SP);
      var dataReader = new DataReader(lines, tagSep + 1);

      ImapData[] parsed = null;

      if (ParseData(dataReader, out parsed, ref requiredBytes)) {
        dataResponseContinuing = false;
        requiredBytes = 0;
        notParsedLines = dataReader.ReadToEnd();
        return ImapDataResponse.Create(parsed);
      }
      else {
        dataResponseContinuing = true;
        requiredBytes += tagSep + 1;
        notParsedLines = lines;
        return null;
      }
    }
  }
}

==> TundereBird-0.28.tar.bz2: TundereBird-0.28/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Protocol.Client/ImapResponseReceiver.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Text;

using Smdn.Formats;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Protocol.Client {
  public class ImapResponseReceiver : ImapReceiver {
    public bool ResponseContinuing {
      get { return dataParsingContext != null; }
    }

    public bool ReceiveResponseAsUTF8 {
      get; set;
    }

    public ImapResponseReceiver(LineOrientedBufferedStream stream)
      : base(stream)
    {
      ReceiveResponseAsUTF8 = false;
    }

    private IParsingContext dataParsingContext;

    public ImapResponse ReceiveResponse()
    {
      var line = Receive();

      try {
        return (dataParsingContext == null)
          ? ParseResponce(line, ref dataParsingContext)
          : ParseDataResponse(line, ref dataParsingContext);
      }
      catch (Exception ex) {
        dataParsingContext = null; // discard context
        if (ex is ImapMalformedDataException)
          throw new ImapMalformedResponseException(ex.Message);
        else
          throw ex;
      }
    }

    private static ByteString continueReqMark = new ByteString("+ ");
    private static ByteString respCondMark = new ByteString("* [");

    private ImapResponse ParseResponce(ByteString line, ref IParsingContext parsingContext)
    {
      // response        = *(continue-req / response-data) response-done

      // continue-req    = "+" SP (resp-text / base64) CRLF
      if (line.StartsWith(continueReqMark))
        return new ImapCommandContinuationRequest(line.Substring(2, line.Length - 4).ToString()); // remove leading "+" SP and trailing CRLF

      // greeting        = "*" SP (resp-cond-auth / resp-cond-bye) CRLF
      // response-done   = response-tagged / response-fatal
      // response-data   = "*" SP (resp-cond-state / resp-cond-bye /
      //                   mailbox-data / message-data / capability-data) CRLF
      // response-fatal  = "*" SP resp-cond-bye CRLF
      //                     ; Server closes connection immediately
      // response-tagged = tag SP resp-cond-state CRLF

      // ("*" / tag) SP
      var tagSep = line.IndexOf(Octets.SP);

      if (tagSep == -1)
        // response-done and response-data must contain SP
        throw new ImapMalformedResponseException("malformed response-done/response-data", line.ToString());

      var untagged = (tagSep == 1 && line[0] == ImapOctets.Asterisk);

      // ("OK" / "BAD" / "NO" / "BYE" / "PREAUTH" / text) SP
      var respCondSep = line.IndexOf(Octets.SP, tagSep + 1);
      var cond = ImapResponseCondition.Undefined;

      if (respCondSep == -1) {
        if (!untagged)
          throw new ImapMalformedResponseException("malformed response-data", line.ToString());
        //else
          // '* SEARCH\r\n' (mailbox-data which contains no SP)
      }
      else {
        cond = ParseCondition(line.Substring(tagSep + 1, respCondSep - tagSep - 1));
      }

      if (cond != ImapResponseCondition.Undefined || line.StartsWith(respCondMark)) {
        // resp-cond-auth / resp-cond-state / resp-cond-bye
        var responseText = ParseRespText((cond == ImapResponseCondition.Undefined)
                                           ? line.Substring(tagSep + 1)
                                           : line.Substring(respCondSep + 1));

        if (untagged)
          return new ImapUntaggedStatusResponse((ImapResponseCondition)cond,
                                            responseText);
        else
          return new ImapTaggedStatusResponse(line.Substring(0, tagSep).ToString(),
                                          (ImapResponseCondition)cond,
                                          responseText);
      }

      // mailbox-data / message-data / capability-data etc.
      return ParseDataResponse(line, ref parsingContext);
    }

    private static ByteString respCondOk      = new ByteString("OK");
    private static ByteString respCondNo      = new ByteString("NO");
    private static ByteString respCondBad     = new ByteString("BAD");
    private static ByteString respCondBye     = new ByteString("BYE");
    private static ByteString respCondPreAuth = new ByteString("PREAUTH");
    
    private ImapResponseCondition ParseCondition(ByteString cond)
    {
      if (respCondOk.EqualsIgnoreCase(cond))
        return ImapResponseCondition.Ok;
      else if (respCondNo.EqualsIgnoreCase(cond))
        return ImapResponseCondition.No;
      else if (respCondBye.EqualsIgnoreCase(cond))
        return ImapResponseCondition.Bye;
      else if (respCondBad.EqualsIgnoreCase(cond))
        return ImapResponseCondition.Bad;
      else if (respCondPreAuth.EqualsIgnoreCase(cond))
        return ImapResponseCondition.PreAuth;
      else
        return ImapResponseCondition.Undefined;
    }

    private ImapResponseText ParseRespText(ByteString respText)
    {
      // resp-text       = ["[" resp-text-code "]" SP] text
      if (respText[0] != ImapOctets.OpenBracket)
        // no resp-text-code
        return new ImapResponseText(DecodeRespTextString(respText.Substring(0, respText.Length - 2))); // remove trailing CRLF

      // exists resp-text-code
      var respTextCodeEnd = respText.IndexOf(ImapOctets.CloseBracket);
      var respTextCode    = respText.Substring(1, respTextCodeEnd - 1);
      var codeSep         = respTextCode.IndexOf(Octets.SP);
      var text            = ByteString.CreateEmpty();

      if (respTextCodeEnd < respText.Length) {
        text = respText.Substring(respTextCodeEnd + 1);
        text = text.Trim(); // remove SP, CRLF
      }

      if (codeSep == -1)
        // no arguments; "READ-ONLY" / "READ-WRITE" / etc...
        return new ImapResponseText(ImapResponseCode.GetKnownOrCreate(respTextCode.ToString()),
                                    new ImapData[] {},
                                    DecodeRespTextString(text));

      // exist arguments; "UIDNEXT" SP nz-number / "UIDVALIDITY" SP nz-number / etc...
      var codeString = respTextCode.Substring(0, codeSep);
      var arguments = ParseDataNonTerminatedText(respTextCode, codeSep + 1);

      if (arguments == null)
        throw new ImapMalformedResponseException("malformed resp-text-code", respTextCode.Substring(codeSep + 1).ToString());

      return new ImapResponseText(ImapResponseCode.GetKnownOrCreate(codeString.ToString(), arguments[0]),
                                  arguments,
                                  DecodeRespTextString(text));
    }

    private string DecodeRespTextString(ByteString text)
    {
      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 3.5. Formal Syntax
       *     resp-text         = ["[" resp-text-code "]" SP ] UTF8-TEXT-CHAR
       *                         *(UTF8-TEXT-CHAR / "[")
       *         ; After the server is changed to a language other than
       *         ; i-default, this resp-text rule replaces the resp-text
       *         ; rule from [RFC3501].
       */
      if (ReceiveResponseAsUTF8)
        return Encoding.UTF8.GetString(text.ByteArray);
      else
        return text.ToString();
    }

    private ImapDataResponse ParseDataResponse(ByteString line, ref IParsingContext parsingContext)
    {
      // mailbox-data / message-data / capability-data etc.

      // ("*" / tag) SP
      var offset = (parsingContext == null) ? line.IndexOf(Octets.SP) + 1 : 0;
      var parsedData = ParseData(line, offset, ref parsingContext);

      if (parsedData == null)
        return null;
      else
        return ImapDataResponse.Create(parsedData);
    }
  }
}
